#include <string.h>
#include "msioGPS.h"
#include "printf.h"
#include "msioLL.h"
#include "msioHL.h"
#include "timer.h"
#include "util.h"


#define REG_DUMMY_READ					0x01
#define REG_RX_FIFO_AVAIL				0x1B
#define REG_RX_FIFO_CTL					0x1E	//bitmask, self-clears, 0x01 - clear sitcks' rx buf, 0x02 - clear tx buf


#define INT_BIT_HAVE_CHAR_TO_GIVE		0x80
#define INT_BIT_HW_RX_OVERFLOW			0x40	//if set, flush stick's tx fifo
#define INT_BIT_PARITY_ERR				0x20	//if set, flush stick's tx fifo
#define INT_BIT_FRAMING_ERR				0x10	//if set, flush stick's tx fifo
#define INT_BIT_ERR_UNK					0x04	//if set, flush stick's rx fifo

struct GpsU16 {
	uint8_t hiBits;	//7 bits
	uint8_t loBits;	//7 bits
} __attribute__((packed));

struct GpsS16 {		//needs sign extension
	uint8_t hiBits;	//7 bits
	uint8_t loBits;	//7 bits
} __attribute__((packed));

struct GpsS32 {
	uint8_t data[4];	//BE, 7 bits each
} __attribute__((packed));


struct PerSatelliteData {
	
	uint8_t satelliteNo;					//0x00
	struct GpsU16 azimuth;					//0x01	satellite position
	uint8_t elevation;						//0x03	satellite position
	uint8_t status;							//0x04	0 = not acquired, 1 = ?, 2 = ? , 3 = ?		todo: sort out
	uint8_t level;							//0x05	signal strength, higher is better
	
} __attribute__((packed));					//0x06

struct GpsDate {
	
	struct GpsU16 year;						//0x01
	uint8_t month;							//0x03
	uint8_t day;							//0x04
	uint8_t hour;							//0x05
	uint8_t minute;							//0x06
	uint8_t second;							//0x07
	
} __attribute__((packed));					//0x07


struct GpsData {
	
	uint8_t val_0xd0;						//0x00
	uint8_t rfu0[0x01];						//0x01
	struct GpsS32 latitude;					//0x02
	struct GpsS32 longitude;				//0x06
	struct GpsS16 altitude;					//0x0a
	struct GpsU16 speed;					//0x0c
	struct GpsU16 heading;					//0x0e
	uint8_t rfu1[0x03];						//0x10
	struct GpsDate currentDate;				//0x13
	uint8_t dayOfWeek;						//0x1a
	struct GpsDate calculatedDate;			//0x1b
	uint8_t rfu2[0x09];						//0x22
	uint8_t calculationMode;				//0x2b	0 = searching, 1,2 = 2d fix, 3 = 3d fix
	uint8_t datum;							//0x2c
	uint8_t rfu3[0x01];						//0x2d
	struct PerSatelliteData satellite[16];	//0x2e
	uint8_t rfu4[0x07];						//0x8e
	uint8_t val_0xda;						//0x95
	
} __attribute__((packed));					//0x96



//XXX: j_GpsDatumToAscii() can be used to see where data is accessed
//XXX: SatelliteCtrl_DrawCircle draws sats - see what it knows
//LonLat_Update

static uint_fast16_t gpsPrvReadU16(const struct GpsU16 *val)
{
	return (((uint_fast16_t)val->hiBits) << 7) + val->loBits;
}

static int_fast16_t gpsPrvReadS16(const struct GpsS16 *val)
{
	uint_fast16_t v = (((uint_fast16_t)val->hiBits) << 7) + val->loBits;
	
	if (v & 0x2000)
		v -= 0x4000;
	
	return v;
}

static int32_t gpsPrvReadS32(const struct GpsS32 *val)
{
	uint_fast8_t i;
	uint32_t v = 0;
	
	for (i = 0; i < 4; i++)
		v = (v << 7) + val->data[i];
	
	if (v & 0x08000000)
		v -= 0x10000000;

	return v;
}

static void gpsPrvShowDate(const struct GpsDate *date)
{
	static const char months[][4] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
	uint_fast16_t year = gpsPrvReadU16(&date->year);
	
	if (!year || !date->month)
		pr("<UNKNOWN>\n");
	else
		pr("%u %s, %u, %02u:%02u:%02u\n", date->day, months[date->month - 1], year, date->hour, date->minute, date->second);
}

static void gpsDecode(struct GpsData *data)
{
	static const char dow[][4] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", };
	bool latNeg = true, lngNeg = true;
	uint_fast16_t heading;
	int32_t lat, lng;
	uint_fast8_t i;
	
	
	pr("\n");
	if (sizeof(struct GpsData) != 0x96) {
		pr("bad size: %u\n", sizeof(struct GpsData));
		while(1);
	}
	
	lat = gpsPrvReadS32(&data->latitude);
	lng = gpsPrvReadS32(&data->longitude);
	heading = gpsPrvReadU16(&data->heading);
	
	if (lat >= 0)
		latNeg = false;
	else
		lat = -lat;
	
	if (lng >= 0)
		lngNeg = false;
	else
		lng = -lng;
	
	pr(" COORDS (Lat: %u*%02u'%02u.%02u\" %c, Long: %u*%02u'%02u.%02u\" %c, Alt: %d m)\n",
		lat / 360000, lat % 360000 / 6000, lat % 6000 / 100, lat % 100, latNeg ? 'S' : 'N',
		lng / 360000, lng % 360000 / 6000, lng % 6000 / 100, lng % 100, lngNeg ? 'W' : 'E',
		gpsPrvReadS16(&data->altitude));
	pr(" MOVEMENT (Speed: %u km/h, Heading: %u.%u degrees)\n", gpsPrvReadU16(&data->speed), heading / 10, heading % 10);
	pr(" DATUM: ");
	switch (data->datum) {
		case 0:
			pr("WGS84\n");
			break;
		
		case 2:
			pr("Tokyo\n");
			break;
		
		case 26:
			pr("ITRF\n");
			break;
		
		default:
			pr("unknown 0x%02x\n", data->datum);
			break;
	}
	pr(" CALC MODE: %s\n", data->calculationMode ? (data->calculationMode < 3 ? "2D FIX" : "3D FIX") : "SEARCHING");
	pr(" DAY: %s\n", dow[data->dayOfWeek]);
	pr(" DATES: current: ");
	gpsPrvShowDate(&data->currentDate);
	pr(" DATES: calced:  ");
	gpsPrvShowDate(&data->calculatedDate);
	pr(" SATELLITES:\n");
	for (i = 0; i < sizeof(data->satellite) / sizeof(*data->satellite); i++) {
		
		pr("  %2u: %3u %5u %u %u %u\n",
			i, data->satellite[i].satelliteNo,
			gpsPrvReadU16(&data->satellite[i].azimuth), 
			data->satellite[i].elevation, data->satellite[i].status, data->satellite[i].level);
	}
}

void msioGpsDemo(void)
{
	static const uint8_t initSeq[] = {0x18, 0x82, 0x1e, 0x03, 0x00, 0x19, 0x03, 0x00, 0x1b, 0x01, 0x1c, 0x01, 0x00, 0x1e, 0x03};
	uint8_t runningBuf[0x96];
	struct ShowBufState sbs;
	uint32_t totalRxed = 0;
	uint_fast8_t i;
	uint8_t buf;
	
	pr("GPS DEMO\n");
	
	if (!msioHlSetRegAddrs(8, 1, 8, 1)) {
		pr("failed to set reg addrs 4\n");
		return;
	}
	
	if (!msioHlRegsRead(&buf, 1)) {
		pr("failed to read reg_0x08\n");
		return;
	}
	
	pr("Needed current: %umA\n", 10 * buf);
	
	//disable function
	buf = 0x00;
	if (!msioHlRegsWrite(&buf, 1)) {
		pr("failed to write regs\n");
		return;
	}
	
	pr("wait a bit...\n");
	delayMsec(50);
	pr("ok, going on\n");
	
	//enable function
	buf = 0xc0;
	if (!msioHlRegsWrite(&buf, 1)) {
		pr("failed to write regs\n");
		return;
	}
	pr("wait a bit...\n");
	delayMsec(100);
	pr("ok, going on\n");
	
	//init
	for (i = 0; i < sizeof(initSeq); i++) {
		
		if (initSeq[i]) {
			if (!msioHlSingleRegWrite(initSeq[i], initSeq[i + 1])) {
				pr("Failed to write %02x to %02x\n", initSeq[i + 1], initSeq[i]);
				return;
			}
			i++;
		}
		else if (!msioHlSingleRegRead(REG_DUMMY_READ, &buf)) {
			pr("Failed to read dummy reg\n");
			return;
		}
	}
	
	pr("we are allegedly inited\n");
	
	while (1) {
		
		if (!msioHlCmdEndWaitInt(TIMER_TICKS_PER_SEC, &buf)) {
			pr("failed to get bytes in a whole second\n");
			break;
		}
	
		if (buf & INT_BIT_HAVE_CHAR_TO_GIVE) {
			if (!msioLlTpcR(MS_RD_LDATA, &buf, 1)) {
				
				pr("failed to read data\n");
				break;
			}
			totalRxed++;
			
			memmove(runningBuf, runningBuf + 1, sizeof(runningBuf) - 1);
			runningBuf[sizeof(runningBuf) - 1] = buf;
			if (runningBuf[0] == 0xd0 && buf == 0xda) {
				
				showBuf(runningBuf, sizeof(runningBuf));
				gpsDecode((void*)runningBuf);
				
				if (0) if (totalRxed >= 2048 && totalRxed < 2048 + 0x96) {
					
					pr("\n\nREQUESTING A COLD START\n");
					
					buf = 0xA5;
					if (!msioLlTpcW(MS_WR_LDATA, &buf, 1)) {
						
						pr("failed to w1\n");
						break;
					}
					buf = 0xDA;
					if (!msioLlTpcW(MS_WR_LDATA, &buf, 1)) {
						
						pr("failed to w2\n");
						break;
					}
					pr("\nREQUESTED A COLD START\n\n");
				}
			}
		} 
	}
	
	while(1);
}


